from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class TextStep(BaseStep):
    name: Literal['text'] = 'text'
    text: str
    new_column: ColumnName


class TextStepWithVariable(TextStep, StepWithVariablesMixin):
    ...



from typing import TypeVar

from pypika.dialects import RedshiftQuery

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import DataTypeMapping, SQLTranslator
from weaverbird.backends.pypika_translator.translators.postgresql import PostgreSQLTranslator

Self = TypeVar("Self", bound="SQLTranslator")


class RedshiftTranslator(PostgreSQLTranslator):
    DIALECT = SQLDialect.REDSHIFT
    QUERY_CLS = RedshiftQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.SIMILAR_TO
    TO_DATE_OP = ToDateOp.TO_DATE


SQLTranslator.register(RedshiftTranslator)



from typing import Any, Dict, List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import EvolutionStep


def translate_evolution(step: EvolutionStep) -> List[MongoStep]:
    new_column = (
        step.new_column
        if step.new_column
        else f'{step.value_col}_EVOL_{step.evolution_format.upper()}'
    )
    error_msg = 'Error: More than one previous date found for the specified index columns'
    add_field_result: Dict[str, Any] = {}

    if step.evolution_format == 'abs':
        add_field_result[new_column] = {
            '$cond': [
                {'$eq': ['$_VQB_VALUE_PREV', 'Error']},
                error_msg,
                {'$subtract': [f'${step.value_col}', '$_VQB_VALUE_PREV']},
            ],
        }
    else:
        add_field_result[new_column] = {
            '$switch': {
                'branches': [
                    {'case': {'$eq': ['$_VQB_VALUE_PREV', 'Error']}, 'then': error_msg},
                    {'case': {'$eq': ['$_VQB_VALUE_PREV', 0]}, 'then': None},
                ],
                'default': {
                    '$divide': [
                        {'$subtract': [f'${step.value_col}', '$_VQB_VALUE_PREV']},
                        '$_VQB_VALUE_PREV',
                    ],
                },
            },
        }

    if step.evolution_type == 'vsLastYear':
        unit = 'year'
    elif step.evolution_type == 'vsLastMonth':
        unit = 'month'
    else:
        unit = 'week'

    return [
        {
            '$addFields': {
                '_VQB_DATE_PREV': {
                    # NOTE: This requires mongo>=5.0
                    '$dateSubtract': {'startDate': f'${step.date_col}', 'unit': unit, 'amount': 1}
                }
            }
        },
        {
            '$facet': {
                '_VQB_ORIGINALS': [{'$project': {'_id': 0}}],
                '_VQB_COPIES_ARRAY': [
                    {'$group': {'_id': None, '_VQB_ALL_DOCS': {'$push': '$$ROOT'}}}
                ],
            },
        },
        {'$unwind': '$_VQB_ORIGINALS'},
        {
            '$project': {
                '_VQB_ORIGINALS': {
                    '$mergeObjects': [
                        '$_VQB_ORIGINALS',
                        {'$arrayElemAt': ['$_VQB_COPIES_ARRAY', 0]},
                    ],
                },
            },
        },
        {'$replaceRoot': {'newRoot': '$_VQB_ORIGINALS'}},
        {
            '$addFields': {
                '_VQB_ALL_DOCS': {
                    '$filter': {
                        'input': '$_VQB_ALL_DOCS',
                        'as': 'item',
                        'cond': {
                            '$and': [
                                {'$eq': ['$_VQB_DATE_PREV', f'$$item.{step.date_col}']},
                                {'$eq': [f'${col}', f'$$item.{col}'] for col in step.index_columns},
                            ],
                        },
                    },
                },
            },
        },
        {
            '$addFields': {
                '_VQB_VALUE_PREV': {
                    '$cond': [
                        {'$gt': [{'$size': f'$_VQB_ALL_DOCS.{step.value_col}'}, 1]},
                        'Error',
                        {'$arrayElemAt': [f'$_VQB_ALL_DOCS.{step.value_col}', 0]},
                    ],
                },
            },
        },
        {'$addFields': add_field_result},
        {
            '$project': {
                '_VQB_ALL_DOCS': 0,
                '_VQB_DATE_PREV': 0,
                '_VQB_VALUE_PREV': 0,
            },
        },
    ]



from typing import Literal, Union

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.combination import Reference


class DomainStep(BaseStep):
    name: Literal['domain'] = 'domain'
    domain: Union[str, Reference]



from typing import TYPE_CHECKING

from pypika import Field, functions
from pypika.dialects import PostgreSQLQuery

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    Self,
    SQLTranslator,
    StepTable,
)

if TYPE_CHECKING:
    from pypika.queries import QueryBuilder

    from weaverbird.pipeline.steps import DurationStep


class PostgreSQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.POSTGRESQL
    QUERY_CLS = PostgreSQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.SIMILAR_TO
    TO_DATE_OP = ToDateOp.TO_DATE

    def duration(
        self: Self, *, step: "DurationStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            functions.Extract(
                step.duration_in, Field(step.end_date_column) - Field(step.start_date_column)
            ).as_(step.new_column_name),
        )
        return query, StepTable(columns=[*table.columns, step.new_column_name])


SQLTranslator.register(PostgreSQLTranslator)



from typing import TYPE_CHECKING, TypeVar

from pypika import Field, Table, functions
from pypika.dialects import MySQLQuery

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    SQLTranslator,
    StepTable,
)

Self = TypeVar("Self", bound="MySQLTranslator")


if TYPE_CHECKING:
    from pypika.queries import QueryBuilder

    from weaverbird.pipeline.steps import SplitStep


class MySQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.MYSQL
    QUERY_CLS = MySQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="SIGNED",
        date="DATE",
        float="DECIMAL",
        integer="UNSIGNED",
        text="CHAR",
    )
    SUPPORT_ROW_NUMBER = False
    SUPPORT_SPLIT_PART = False
    FROM_DATE_OP = FromDateOp.DATE_FORMAT
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.STR_TO_DATE

    def split(
        self: Self, *, step: "SplitStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.column]
        new_cols = [f"{step.column}_{i+1}" for i in range(step.number_cols_to_keep)]

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            *(
                # https://stackoverflow.com/a/32500349
                SubstringIndex(
                    SubstringIndex(col_field, step.delimiter, i + 1), step.delimiter, -1
                ).as_(new_cols[i])
                for i in range(step.number_cols_to_keep)
            ),
        )
        return query, StepTable(columns=[*table.columns, *new_cols])


SQLTranslator.register(MySQLTranslator)


class SubstringIndex(functions.Function):  # type: ignore[misc]
    def __init__(
        self, term: str | Field, delimiter: str, count: int, alias: str | None = None
    ) -> None:
        super().__init__("SUBSTRING_INDEX", term, delimiter, count, alias=alias)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import LowercaseStep


def translate_lowercase(step: LowercaseStep) -> List[MongoStep]:
    return [{'$addFields': {step.column: {'$toLower': f'${step.column}'}}}]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.cumsum import CumSumStep


def translate_cumsum(step: CumSumStep) -> List[MongoStep]:
    groupby = step.groupby or []

    return [
        {'$sort': {step.reference_column: 1}},
        {
            '$group': {
                '_id': {col: f'${col}' for col in groupby} if step.groupby else None,
                **{x: {'$push': f'${x}'} for x, _ in step.to_cumsum},
                '_vqbArray': {'$push': '$$ROOT'},
            },
        },
        {'$unwind': {'path': '$_vqbArray', 'includeArrayIndex': '_VQB_INDEX'}},
        {
            '$project': {
                **{col: f'$_id.{col}' for col in groupby},
                **{
                    new_name
                    if new_name
                    else f'{name}_CUMSUM': {
                        '$sum': {'$slice': [f'${name}', {'$add': ['$_VQB_INDEX', 1]}]}
                    }
                    for name, new_name in step.to_cumsum
                },
                '_vqbArray': 1,
            },
        },
        {'$replaceRoot': {'newRoot': {'$mergeObjects': ['$_vqbArray', '$$ROOT']}}},
        {'$project': {'_vqbArray': 0}},
    ]



from typing import Literal

from pydantic import validator

from weaverbird.pipeline.steps.utils.base import BaseStep


class HierarchyStep(BaseStep):
    name: Literal['hierarchy'] = 'hierarchy'
    hierarchy_level_column: str = 'hierarchy_level'
    hierarchy: list[str]
    include_nulls: bool = False

    @validator('hierarchy')
    def _ensure_hierarchy(cls, values: list[str]) -> list[str]:
        assert len(values) > 0, "at least one value must be specified"
        return values



import ast
import re
from typing import Any, Dict, Tuple, Union

from weaverbird.pipeline.steps import FormulaStep


def mongo_formula_for_constant(constant: ast.Constant) -> Any:
    return constant.value


def mongo_formula_for_name(name: ast.Name, columns_aliases: Dict) -> str:
    """
    Identifiers correspond to column names
    """
    if name.id in columns_aliases:
        return f'${columns_aliases[name.id]}'
    else:
        return '$' + name.id


def mongo_formula_for_binop(binop: ast.BinOp, columns_aliases: Dict) -> dict:
    if isinstance(binop.op, ast.Add):
        mongo_op = '$add'
    elif isinstance(binop.op, ast.Sub):
        mongo_op = '$subtract'
    elif isinstance(binop.op, ast.Mult):
        mongo_op = '$multiply'
    elif isinstance(binop.op, ast.Div):
        mongo_op = '$divide'
    elif isinstance(binop.op, ast.Pow):
        mongo_op = '$pow'
    else:
        raise InvalidFormula(f'Operator {binop.op.__class__} is not supported')

    translated_op = {
        mongo_op: [
            mongo_formula_for_ast_node(binop.left, columns_aliases),
            mongo_formula_for_ast_node(binop.right, columns_aliases),
        ]
    }
    #  In case of a division operation the translator must handle a 0 or null denominator.
    #  The implemented logic is: if the translated operation is a division, wraps the operation with a
    #  condition checking if the value of the right operand is in 0 or None and output a None result in this case
    #  otherwise, return the result the division operation
    if mongo_op == '$divide':
        return {
            '$cond': [
                {'$in': [mongo_formula_for_ast_node(binop.right, columns_aliases), [0, None]]},
                None,
                translated_op,
            ]
        }
    else:
        return translated_op


def mongo_formula_for_unaryop(unop: ast.UnaryOp, columns_aliases: Dict) -> dict:
    if isinstance(unop.op, ast.USub):
        return {'$multiply': [-1, mongo_formula_for_ast_node(unop.operand, columns_aliases)]}
    else:
        raise InvalidFormula(f'Operator {unop.op.__class__} is not supported')


def mongo_formula_for_ast_node(node: ast.AST, columns_aliases: Dict) -> Any:
    if isinstance(node, ast.BinOp):
        return mongo_formula_for_binop(node, columns_aliases)
    elif isinstance(node, ast.UnaryOp):
        return mongo_formula_for_unaryop(node, columns_aliases)
    elif isinstance(node, ast.Constant):
        return mongo_formula_for_constant(node)
    elif isinstance(node, ast.Name):
        return mongo_formula_for_name(node, columns_aliases)
    else:
        raise InvalidFormula(f'Formula node {node} is not supported')


def mongo_formula_for_expr(expr: ast.Expr, columns_aliases: Dict) -> Dict[str, Any]:
    if isinstance(expr.value, ast.AST):
        return mongo_formula_for_ast_node(expr.value, columns_aliases)
    else:
        raise InvalidFormula


def translate_formula(step: FormulaStep) -> list:
    return [{'$addFields': {step.new_column: build_mongo_formula_tree(step.formula)}}]


def build_mongo_formula_tree(
    formula: Union[str, int, float, bool]
) -> Union[Dict[str, Any], int, float, bool]:
    if isinstance(formula, str):
        columns_aliases, sanitized_formula = sanitize_formula(formula)
        module = ast.parse(sanitized_formula)
        expr = module.body[0]
        if not isinstance(expr, ast.Expr):
            raise InvalidFormula
        mongo_expr = mongo_formula_for_expr(expr, columns_aliases)
        return mongo_expr
    elif type(formula) in (int, float, bool):
        return formula
    else:
        raise InvalidFormula


def sanitize_formula(formula: str) -> Tuple[Dict[str, str], str]:
    """
    This function handles column names with special characters & spaces.
    Such columns are surrounded by brackets. The functions replaces this pattern by a temporary name
    and store the original name in a dict, with the temporary name as key and the actual name as key.
    For example in this formula -> [I'm a Special Column!!] * 10 it will return
    tuple(__vqb_col_0__ * 10, {"__vqb_col_0__": "[I'm a Special Column!!]"})
    The translate function will replaces the temporary name by the old one once the formula will
    be parsed by ast.parse and translated to mongo.
    """
    columns_aliases = {}
    sanitized_formula = formula
    matches = re.findall(r'\[(.*?)]', formula)
    for i, match in enumerate(matches):
        columns_aliases[f'__vqb_col_{i}__'] = match
        sanitized_formula = sanitized_formula.replace(f'[{match}]', f'__vqb_col_{i}__')
    return columns_aliases, sanitized_formula


class InvalidFormula(Exception):
    """
    Raised when a formula is not supported
    """



from enum import Enum, auto


class RegexOp(Enum):
    REGEXP = auto()
    SIMILAR_TO = auto()
    CONTAINS = auto()


class FromDateOp(Enum):
    DATE_FORMAT = auto()
    TO_CHAR = auto()


class ToDateOp(Enum):
    TO_DATE = auto()
    STR_TO_DATE = auto()
    PARSE_DATE = auto()



from typing import List, Literal, Optional, Sequence

from pydantic import Field, root_validator, validator
from pydantic.main import BaseModel

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.steps.utils.validation import validate_unique_columns
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames, TemplatedVariable

AggregateFn = Literal[
    'avg',
    'sum',
    'min',
    'max',
    'count',
    'count distinct',
    'first',
    'last',
    'count distinct including empty',
]


class Aggregation(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    new_columns: List[ColumnName] = Field(alias='newcolumns')
    agg_function: AggregateFn = Field(alias='aggfunction')
    columns: List[ColumnName]

    @validator('columns', pre=True)
    def validate_unique_columns(cls, value):
        return validate_unique_columns(value)

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'column' in values:
            values['columns'] = [values.pop('column')]
        if 'newcolumn' in values:
            values['new_columns'] = [values.pop('newcolumn')]
        return values


class AggregateStep(BaseStep):
    name: Literal['aggregate'] = 'aggregate'
    on: List[ColumnName] = []
    aggregations: Sequence[Aggregation]
    keep_original_granularity: Optional[bool] = Field(
        default=False, alias='keepOriginalGranularity'
    )

    class Config(PopulatedWithFieldnames):
        ...


class AggregationWithVariables(Aggregation):
    class Config(PopulatedWithFieldnames):
        ...

    new_columns: List[TemplatedVariable] = Field(alias='newcolumns')
    agg_function: TemplatedVariable = Field(alias='aggfunction')
    columns: List[TemplatedVariable]


class AggregateStepWithVariables(AggregateStep, StepWithVariablesMixin):
    aggregations: Sequence[AggregationWithVariables]



from abc import ABC
from dataclasses import dataclass

# from typing_extensions import Self
from typing import TYPE_CHECKING, Any, Callable, Mapping, Sequence, TypeVar, cast

from pypika import AliasedQuery, Case, Criterion, Field, Order, Query, Schema, Table, functions
from pypika.enums import Comparator
from pypika.terms import AnalyticFunction, BasicCriterion, LiteralValue

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators import ALL_TRANSLATORS
from weaverbird.exceptions import MissingTableNameError
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.steps.utils.combination import Reference

Self = TypeVar("Self", bound="SQLTranslator")


if TYPE_CHECKING:
    from pypika.queries import QueryBuilder

    from weaverbird.pipeline import PipelineStep
    from weaverbird.pipeline.conditions import Condition, SimpleCondition
    from weaverbird.pipeline.steps import (
        AggregateStep,
        ArgmaxStep,
        ArgminStep,
        CompareTextStep,
        ConcatenateStep,
        ConvertStep,
        CustomSqlStep,
        DeleteStep,
        DomainStep,
        DuplicateStep,
        FillnaStep,
        FilterStep,
        FormulaStep,
        FromdateStep,
        IfthenelseStep,
        LowercaseStep,
        PercentageStep,
        RenameStep,
        ReplaceStep,
        SelectStep,
        SortStep,
        SplitStep,
        SubstringStep,
        TextStep,
        ToDateStep,
        TopStep,
        TrimStep,
        UniqueGroupsStep,
        UppercaseStep,
    )
    from weaverbird.pipeline.steps.aggregate import AggregateFn


@dataclass(kw_only=True)
class StepTable:
    columns: list[str]
    name: str | None = None


@dataclass(kw_only=True)
class DataTypeMapping:
    boolean: str
    date: str
    float: str
    integer: str
    text: str


class SQLTranslator(ABC):
    DIALECT: SQLDialect
    QUERY_CLS: Query
    DATA_TYPE_MAPPING: DataTypeMapping
    # supported extra functions
    SUPPORT_ROW_NUMBER: bool
    SUPPORT_SPLIT_PART: bool
    # which operators should be used
    FROM_DATE_OP: FromDateOp
    REGEXP_OP: RegexOp
    TO_DATE_OP: ToDateOp

    def __init__(
        self: Self,
        *,
        tables_columns: Mapping[str, Sequence[str]] | None = None,
        db_schema: str | None = None,
    ) -> None:
        self._tables_columns: Mapping[str, Sequence[str]] = tables_columns or {}
        self._db_schema: Schema | None = Schema(db_schema) if db_schema is not None else None

    def __init_subclass__(cls) -> None:
        ALL_TRANSLATORS[cls.DIALECT] = cls

    def get_query(self: Self, *, steps: Sequence["PipelineStep"]) -> "QueryBuilder":
        step_queries: list["QueryBuilder"] = []
        step_tables: list[StepTable] = []

        for i, step in enumerate(steps):
            try:
                step_method: Callable[..., tuple["QueryBuilder", StepTable]] = getattr(
                    self, step.name
                )
            except AttributeError:
                raise NotImplementedError(f"[{self.DIALECT}] step {step.name} not yet implemented")

            if i == 0:
                assert step.name == "domain"
                step_query, step_table = step_method(step=step)
            else:
                step_query, step_table = step_method(step=step, table=step_tables[i - 1])

            step_queries.append(step_query)
            step_table.name = f"__step_{i}__"
            step_tables.append(step_table)

        query: "QueryBuilder" = self.QUERY_CLS
        for i, step_query in enumerate(step_queries):
            query = query.with_(step_query, step_tables[i].name)

        return query.from_(step_tables[-1].name).select("*")

    def get_query_str(self: Self, *, steps: Sequence["PipelineStep"]) -> str:
        query_str: str = self.get_query(steps=steps).get_sql()
        return query_str

    # All other methods implement step from https://weaverbird.toucantoco.com/docs/steps/,
    # the name of the method being the name of the step and the kwargs the rest of the params
    def _get_aggregate_function(
        self: Self, agg_function: "AggregateFn"
    ) -> functions.AggregateFunction:
        match agg_function:
            case "avg":
                return functions.Avg
            case "count":
                return functions.Count
            case "count distinct":
                return CountDistinct
            case "max":
                return functions.Max
            case "min":
                return functions.Min
            case "sum":
                return functions.Sum
            case _:  # pragma: no cover
                raise NotImplementedError(
                    f"[{self.DIALECT}] Aggregation for {agg_function!r} is not yet implemented"
                )

    def aggregate(
        self: Self, *, step: "AggregateStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        the_table = Table(table.name)
        agg_selected: list[Field] = []

        for aggregation in step.aggregations:
            agg_fn = self._get_aggregate_function(aggregation.agg_function)
            for i, column_name in enumerate(aggregation.columns):
                column_field: Field = the_table[column_name]
                new_agg_col = agg_fn(column_field).as_(aggregation.new_columns[i])
                agg_selected.append(new_agg_col)

        query: "QueryBuilder"
        selected_col_names: list[str]

        if step.keep_original_granularity:
            current_query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*table.columns)

            agg_query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
                *step.on, *agg_selected
            )
            agg_query = agg_query.groupby(*step.on)

            all_agg_col_names: list[str] = [x for agg in step.aggregations for x in agg.new_columns]

            query = (
                self.QUERY_CLS.from_(current_query)
                .select(
                    *table.columns,
                    *(Field(agg_col, table=agg_query) for agg_col in all_agg_col_names),
                )
                .left_join(agg_query)
                .on_field(*step.on)
            )
            selected_col_names = [*table.columns, *all_agg_col_names]

        else:
            selected_cols: list[str | Field] = [*step.on, *agg_selected]
            selected_col_names = [*step.on, *(f.alias for f in agg_selected)]
            query = (
                self.QUERY_CLS.from_(table.name)
                .select(*selected_cols)
                .groupby(*step.on)
                .orderby(*step.on, order=Order.asc)
            )

        return query, StepTable(columns=selected_col_names)

    def argmax(
        self: Self, *, step: "ArgmaxStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        from weaverbird.pipeline.steps import TopStep

        return self.top(
            step=TopStep(rank_on=step.column, sort="desc", limit=1, groups=step.groups), table=table
        )

    def argmin(
        self: Self, *, step: "ArgminStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        from weaverbird.pipeline.steps import TopStep

        return self.top(
            step=TopStep(rank_on=step.column, sort="asc", limit=1, groups=step.groups), table=table
        )

    def comparetext(
        self: Self, *, step: "CompareTextStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            Case()
            .when(Table(table.name)[step.str_col_1] == Table(table.name)[step.str_col_2], True)
            .else_(False)
            .as_(step.new_column_name),
        )
        return query, StepTable(columns=[*table.columns, step.new_column_name])

    def concatenate(
        self: Self, *, step: "ConcatenateStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        # from step.columns = ["city", "age", "username"], step.separator = " -> "
        # create [Field("city"), " -> ", Field("age"), " -> ", Field("username")]
        the_table = Table(table.name)
        tokens = [the_table[step.columns[0]]]
        for col in step.columns[1:]:
            tokens.append(step.separator)
            tokens.append(the_table[col])

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            functions.Concat(*tokens).as_(step.new_column_name),
        )
        return query, StepTable(columns=[*table.columns, step.new_column_name])

    def convert(
        self: Self, *, step: "ConvertStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_fields: list[Field] = [Table(table.name)[col] for col in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c not in step.columns),
            *(
                functions.Cast(col_field, getattr(self.DATA_TYPE_MAPPING, step.data_type)).as_(
                    col_field.name
                )
                for col_field in col_fields
            ),
        )
        return query, StepTable(columns=table.columns)

    def customsql(
        self: Self, *, step: "CustomSqlStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        """create a custom sql step based on the current table named ##PREVIOUS_STEP## in the query"""

        class CustomQuery(AliasedQuery):  # type: ignore[misc]
            def get_sql(self, **kwargs: Any) -> str:
                return cast(str, self.query)

        if table.name is None:
            raise MissingTableNameError()

        custom_query = CustomQuery(
            name=f"custom_from_{table.name}",
            query=step.query.replace("##PREVIOUS_STEP##", table.name),
        )

        # we now have no way to know which columns remain
        # without actually executing the query
        return custom_query, StepTable(columns=["*"])

    def delete(
        self: Self, *, step: "DeleteStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        new_columns = [c for c in table.columns if c not in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*new_columns)
        return query, StepTable(columns=new_columns)

    def domain(
        self: Self,
        *,
        step: "DomainStep",
    ) -> tuple["QueryBuilder", StepTable]:
        try:
            if isinstance(step.domain, Reference):
                raise NotImplementedError(f"[{self.DIALECT}] Cannot resolve a reference to a query")
            else:
                selected_cols = list(self._tables_columns[step.domain])
        except KeyError:
            selected_cols = ["*"]

        query: "QueryBuilder" = self.QUERY_CLS.from_(
            Table(step.domain, schema=self._db_schema)
        ).select(*selected_cols)
        return query, StepTable(columns=selected_cols)

    def duplicate(
        self: Self, *, step: "DuplicateStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns, Table(table.name)[step.column].as_(step.new_column_name)
        )
        return query, StepTable(columns=[*table.columns, step.new_column_name])

    def fillna(
        self: Self, *, step: "FillnaStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        the_table = Table(table.name)
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c not in step.columns),
            *(
                functions.Coalesce(the_table[col_name], step.value).as_(col_name)
                for col_name in step.columns
            ),
        )
        return query, StepTable(columns=table.columns)

    def _get_single_condition_criterion(
        self: Self, condition: "SimpleCondition", table: StepTable
    ) -> Criterion:
        column_field: Field = Table(table.name)[condition.column]

        match condition:
            case ComparisonCondition():
                import operator

                op = getattr(operator, condition.operator)
                return op(column_field, condition.value)

            case InclusionCondition():
                if condition.operator == "in":
                    return column_field.isin(condition.value)
                elif condition.operator == "nin":
                    return column_field.notin(condition.value)

            case MatchCondition():

                if condition.operator == "matches":
                    match self.REGEXP_OP:
                        case RegexOp.REGEXP:
                            return column_field.regexp(condition.value)
                        case RegexOp.SIMILAR_TO:
                            return BasicCriterion(
                                RegexpMatching.similar_to,
                                column_field,
                                column_field.wrap_constant(_compliant_regex(condition.value)),
                            )
                        case RegexOp.CONTAINS:
                            return BasicCriterion(
                                RegexpMatching.contains,
                                column_field,
                                column_field.wrap_constant(_compliant_regex(condition.value)),
                            )
                        case _:
                            raise NotImplementedError(
                                f"[{self.DIALECT}] doesn't have regexp operator"
                            )

                elif condition.operator == "notmatches":
                    match self.REGEXP_OP:
                        case RegexOp.REGEXP:
                            return column_field.regexp(condition.value).negate()
                        case RegexOp.SIMILAR_TO:
                            return BasicCriterion(
                                RegexpMatching.not_similar_to,
                                column_field,
                                column_field.wrap_constant(_compliant_regex(condition.value)),
                            )
                        case RegexOp.CONTAINS:
                            return BasicCriterion(
                                RegexpMatching.not_contains,
                                column_field,
                                column_field.wrap_constant(_compliant_regex(condition.value)),
                            )
                        case _:
                            raise NotImplementedError(
                                f"[{self.DIALECT}] doesn't have regexp operator"
                            )

            case NullCondition():
                if condition.operator == "isnull":
                    return column_field.isnull()
                elif condition.operator == "notnull":
                    return column_field.isnotnull()

            case DateBoundCondition():
                if condition.operator == "from":
                    return column_field <= condition.value
                elif condition.operator == "until":
                    return column_field >= condition.value

            case _:  # pragma: no cover
                raise KeyError(f"Operator {condition.operator!r} does not exist")

    def _get_filter_criterion(self: Self, condition: "Condition", table: StepTable) -> Criterion:
        from weaverbird.pipeline.conditions import ConditionComboAnd, ConditionComboOr

        match condition:
            case ConditionComboOr():
                return Criterion.any(
                    (self._get_filter_criterion(condition, table) for condition in condition.or_)
                )
            case ConditionComboAnd():
                return Criterion.all(
                    (self._get_filter_criterion(condition, table) for condition in condition.and_)
                )
            case _:
                return self._get_single_condition_criterion(condition, table)

    def filter(
        self: Self, *, step: "FilterStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = (
            self.QUERY_CLS.from_(table.name)
            .select(*table.columns)
            .where(self._get_filter_criterion(step.condition, table))
        )
        return query, StepTable(columns=table.columns)

    def formula(
        self: Self, *, step: "FormulaStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*table.columns)
        # TODO: support
        # - float casting with divisions
        # - whitespaces in column names
        # - strings
        # To do that we probably need to parse the formula with tokens
        # In the end we should be able to translate:
        #   [my age] + 1 / 2
        # into
        #   CAST("my age" AS float) + 1 / 2
        query = query.select(LiteralValue(step.formula).as_(step.new_column))
        return query, StepTable(columns=[*table.columns, step.new_column])

    def fromdate(
        self: Self, *, step: "FromdateStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.column]

        match self.FROM_DATE_OP:
            case FromDateOp.DATE_FORMAT:
                convert_fn = DateFormat
            case FromDateOp.TO_CHAR:
                convert_fn = functions.ToChar
            case _:
                raise NotImplementedError(f"[{self.DIALECT}] doesn't have from date operator")

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c != step.column),
            convert_fn(col_field, step.format).as_(step.column),
        )
        return query, StepTable(columns=table.columns)

    def _build_ifthenelse_case(
        self,
        *,
        if_: "Condition",
        then_: Any,
        else_: "Condition" | Any,
        table: StepTable,
        case: Case,
    ) -> Case:
        import json

        from weaverbird.pipeline.steps.ifthenelse import IfThenElse

        try:
            # if the value is a string
            then_value = json.loads(then_)
            case = case.when(self._get_filter_criterion(if_, table), then_value)
        except (json.JSONDecodeError, TypeError):
            # the value is a formula
            then_value = then_
            case = case.when(self._get_filter_criterion(if_, table), LiteralValue(then_value))

        if isinstance(else_, IfThenElse):
            return self._build_ifthenelse_case(
                if_=else_.condition,
                then_=else_.then,
                else_=else_.else_value,
                table=table,
                case=case,
            )
        else:
            try:
                # the value is a string
                else_value = json.loads(else_)  # type: ignore
                return case.else_(else_value)
            except (json.JSONDecodeError, TypeError):
                # the value is a formula
                else_value = else_
                return case.else_(LiteralValue(else_value))

    def ifthenelse(
        self: Self, *, step: "IfthenelseStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            self._build_ifthenelse_case(
                if_=step.condition, then_=step.then, else_=step.else_value, table=table, case=Case()
            ).as_(step.new_column),
        )

        return query, StepTable(columns=[*table.columns, step.new_column])

    def lowercase(
        self: Self, *, step: "LowercaseStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c != step.column),
            functions.Lower(col_field).as_(step.column),
        )
        return query, StepTable(columns=table.columns)

    def percentage(
        self: Self, *, step: "PercentageStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        raise NotImplementedError(f"[{self.DIALECT}] percentage is not implemented")

    def rename(
        self: Self, *, step: "RenameStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        new_names_mapping: dict[str, str] = dict(step.to_rename)

        selected_col_fields: list[Field] = []

        for col_name in table.columns:
            if col_name in new_names_mapping:
                selected_col_fields.append(Field(name=col_name, alias=new_names_mapping[col_name]))
            else:
                selected_col_fields.append(Field(name=col_name))

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*selected_col_fields)
        return query, StepTable(columns=[f.alias or f.name for f in selected_col_fields])

    def replace(
        self: Self, *, step: "ReplaceStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.search_column]

        # Do a nested `replace` to replace many values on the same column
        replaced_col = col_field
        for old_name, new_name in step.to_replace:
            replaced_col = functions.Replace(replaced_col, old_name, new_name)

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c != step.search_column),
            replaced_col.as_(step.search_column),
        )

        return query, StepTable(columns=table.columns)

    def select(
        self: Self, *, step: "SelectStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*step.columns)
        return query, StepTable(columns=step.columns)

    def sort(self: Self, *, step: "SortStep", table: StepTable) -> tuple["QueryBuilder", StepTable]:
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*table.columns)

        for column_sort in step.columns:
            query = query.orderby(
                column_sort.column, order=Order.desc if column_sort.order == "desc" else Order.asc
            )

        return query, StepTable(columns=table.columns)

    def split(
        self: Self, *, step: "SplitStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        if self.SUPPORT_SPLIT_PART:
            col_field: Field = Table(table.name)[step.column]
            new_cols = [f"{step.column}_{i+1}" for i in range(step.number_cols_to_keep)]
            query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
                *table.columns,
                *(
                    functions.SplitPart(col_field, step.delimiter, i + 1).as_(new_cols[i])
                    for i in range(step.number_cols_to_keep)
                ),
            )
            return query, StepTable(columns=[*table.columns, *new_cols])

        raise NotImplementedError(f"[{self.DIALECT}] split is not implemented")

    def substring(
        self: Self, *, step: "SubstringStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        step.new_column_name = (
            f"{step.column}_SUBSTR" if step.new_column_name is None else step.new_column_name
        )
        col_field: Field = Table(table.name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns,
            functions.Substring(
                col_field, step.start_index, (step.end_index - step.start_index) + 1
            ).as_(step.new_column_name),
        )
        return query, StepTable(columns=[*table.columns, step.new_column_name])

    def text(self: Self, *, step: "TextStep", table: StepTable) -> tuple["QueryBuilder", StepTable]:
        from pypika.terms import ValueWrapper

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *table.columns, ValueWrapper(step.text).as_(step.new_column)
        )
        return query, StepTable(columns=[*table.columns, step.new_column])

    def todate(
        self: Self, *, step: "ToDateStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.column]

        if step.format is not None:
            match self.TO_DATE_OP:
                case ToDateOp.TO_DATE:
                    convert_fn = functions.ToDate
                case ToDateOp.STR_TO_DATE:
                    convert_fn = StrToDate
                case ToDateOp.PARSE_DATE:
                    convert_fn = ParseDate
                case _:
                    raise NotImplementedError(f"[{self.DIALECT}] todate has no set operator")
            date_selection = convert_fn(col_field, step.format)
        else:
            date_selection = functions.Cast(col_field, self.DATA_TYPE_MAPPING.date)

        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c != step.column),
            date_selection.as_(col_field.name),
        )
        return query, StepTable(columns=table.columns)

    def top(self: Self, *, step: "TopStep", table: StepTable) -> tuple["QueryBuilder", StepTable]:
        if step.groups:
            if self.SUPPORT_ROW_NUMBER:
                sub_query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(*table.columns)

                the_table = Table(table.name)
                rank_on_field: Field = the_table[step.rank_on]
                groups_fields: list[Field] = [the_table[group] for group in step.groups]
                sub_query = sub_query.select(
                    RowNumber()
                    .over(*groups_fields)
                    .orderby(rank_on_field, order=Order.desc if step.sort == "desc" else Order.asc)
                )
                query: "QueryBuilder" = (
                    self.QUERY_CLS.from_(sub_query)
                    .select(*table.columns)
                    .where(Field("row_number") == step.limit)
                )
                return query, StepTable(columns=table.columns)

            else:
                raise NotImplementedError(f"[{self.DIALECT}] top is not implemented with groups")

        query = (
            self.QUERY_CLS.from_(table.name)
            .select(*table.columns)
            .orderby(step.rank_on, order=Order.desc if step.sort == "desc" else Order.asc)
            .limit(step.limit)
        )
        return query, StepTable(columns=table.columns)

    def trim(self: Self, *, step: "TrimStep", table: StepTable) -> tuple["QueryBuilder", StepTable]:
        col_fields: list[Field] = [Table(table.name)[col] for col in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c not in step.columns),
            *(functions.Trim(col_field).as_(col_field.name) for col_field in col_fields),
        )
        return query, StepTable(columns=table.columns)

    def uniquegroups(
        self: Self, *, step: "UniqueGroupsStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        from weaverbird.pipeline.steps import AggregateStep

        return self.aggregate(
            step=AggregateStep(on=step.on, aggregations=[], keepOriginalGranularity=False),
            table=table,
        )

    def uppercase(
        self: Self, *, step: "UppercaseStep", table: StepTable
    ) -> tuple["QueryBuilder", StepTable]:
        col_field: Field = Table(table.name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(table.name).select(
            *(c for c in table.columns if c != step.column),
            functions.Upper(col_field).as_(step.column),
        )
        return query, StepTable(columns=table.columns)


class CountDistinct(functions.Count):  # type: ignore[misc]
    def __init__(self, param: str | Field, alias: str | None = None) -> None:
        super().__init__(param, alias)
        self._distinct = True


class DateFormat(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("DATE_FORMAT", term, date_format, alias=alias)


class RowNumber(AnalyticFunction):  # type: ignore[misc]
    def __init__(self, **kwargs: Any) -> None:
        super().__init__("ROW_NUMBER", **kwargs)


class StrToDate(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("STR_TO_DATE", term, date_format, alias=alias)


class ParseDate(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("PARSE_DATE", term, date_format, alias=alias)


class RegexpMatching(Comparator):  # type: ignore[misc]
    similar_to = " SIMILAR TO "
    not_similar_to = " NOT SIMILAR TO "
    contains = " CONTAINS "
    not_contains = " NOT CONTAINS "


def _compliant_regex(pattern: str) -> str:
    """
    Like LIKE, the SIMILAR TO operator succeeds only if its pattern matches the entire string;
    this is unlike common regular expression behavior wherethe pattern
    can match any part of the string
    (see https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-SIMILARTO-REGEXP)
    """
    return f"%{pattern}%"



from pypika.dialects import Query

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.translators.base import SQLTranslator


class AthenaTranslator(SQLTranslator):
    DIALECT = SQLDialect.ATHENA
    QUERY_CLS = Query


SQLTranslator.register(AthenaTranslator)



from typing import List, Union

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ConvertStep

TYPE_MAP = {
    'boolean': 'bool',
    'date': 'date',
    'float': 'double',
    'integer': 'long',  # better to cast into 64-bit integers. 32-bit integers ('int') cannot be casted to dates
    'text': 'string',
}


def translate_convert(step: ConvertStep) -> List[MongoStep]:
    # /!\ we want to get the same results than with the mongo typescript
    # translator, not than with the pandas executor.

    mongo_add_fields = {}
    mongo_type = TYPE_MAP[step.data_type] if step.data_type in TYPE_MAP else ''
    for column in step.columns:
        # Mongo cannot cast integers into date but only long into date, so we
        # manage the cast from int to long when needed.
        _input: Union[dict, str]
        if mongo_type == 'date':
            _input = {
                '$cond': [
                    {'$eq': [{'$type': f'${column}'}, 'int']},
                    {'$toLong': f'${column}'},
                    f'${column}',
                ]
            }
        else:
            _input = f'${column}'
        mongo_add_fields[column] = {
            '$convert': {
                'input': _input,
                'to': mongo_type,
                'onError': None,  # failed conversions result in null instead of crashing
            }
        }

    return [{'$addFields': mongo_add_fields}]



from typing import List, Literal, Union

from pydantic import BaseModel


class Reference(BaseModel):
    type: Literal['ref']
    uid: str


PipelineOrDomainName = Union[List[dict], str]  # can be either a domain name or a complete pipeline
PipelineOrDomainNameOrReference = Union[PipelineOrDomainName, Reference]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import TopStep


def translate_top(
    step: TopStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'TOP_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    # fields to complete the query
    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str()

    if len(step.groups) > 0:
        order_on_groupby = ' ASC, '.join(step.groups)
        if len(step.groups) > 0:
            order_on_groupby += ' ASC'

        final_query = (
            f"SELECT * FROM (SELECT * FROM {query.query_name} QUALIFY ROW_NUMBER() "
            f"OVER (PARTITION BY {', '.join(step.groups)} ORDER BY {step.rank_on} {step.sort}) <= {step.limit}) "
            f"{query_name}_ALIAS ORDER BY {order_on_groupby}"
        )
    else:
        final_query = (
            f"SELECT {completed_fields} FROM {query.query_name} ORDER BY {step.rank_on}"
            f" {step.sort} LIMIT {step.limit}"
        )

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"{query.transformed_query}, {query_name} AS ({final_query})",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from enum import Enum
from typing import TYPE_CHECKING, Any, TypeVar

from pypika import Criterion, Field, Query, Table, functions
from pypika.queries import QueryBuilder

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    SQLTranslator,
    StepTable,
)
from weaverbird.pipeline.conditions import DateBoundCondition

Self = TypeVar("Self", bound="GoogleBigQueryTranslator")

if TYPE_CHECKING:
    from weaverbird.pipeline.conditions import SimpleCondition


class ExtraDialects(Enum):
    GOOGLE_BIG_QUERY = "googlebigquery"


class GoogleBigQueryQuery(Query):  # type: ignore[misc]
    @classmethod
    def _builder(cls, **kwargs: Any) -> "GoogleBigQueryQueryBuilder":
        return GoogleBigQueryQueryBuilder(**kwargs)


class GoogleBigQueryQueryBuilder(QueryBuilder):  # type: ignore[misc]
    QUOTE_CHAR = "`"
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = None
    QUERY_CLS = GoogleBigQueryQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=ExtraDialects.GOOGLE_BIG_QUERY, **kwargs)


class GoogleBigQueryTranslator(SQLTranslator):
    DIALECT = SQLDialect.GOOGLEBIGQUERY
    QUERY_CLS = GoogleBigQueryQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = False
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.CONTAINS
    TO_DATE_OP = ToDateOp.PARSE_DATE

    def _get_single_condition_criterion(
        self: Self, condition: "SimpleCondition", table: StepTable
    ) -> Criterion:
        column_field: Field = Table(table.name)[condition.column]

        if isinstance(condition, DateBoundCondition):
            match condition.operator:
                case "from":
                    return functions.Cast(column_field, "datetime") >= ParseDatetime(
                        "%FT%T", condition.value
                    )
                case "until":
                    return functions.Cast(column_field, "datetime") <= ParseDatetime(
                        "%FT%T", condition.value
                    )

        return super()._get_single_condition_criterion(condition, table)


SQLTranslator.register(GoogleBigQueryTranslator)


class ParseDatetime(functions.Function):  # type: ignore[misc]
    def __init__(self, format: str, term: str | Field, alias: str | None = None) -> None:
        super().__init__("parse_datetime", format, term, alias=alias)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ConcatenateStep


def execute_concatenate(
    step: ConcatenateStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_col = df[step.columns[0]].astype(str)
    for col_name in step.columns[1:]:
        new_col = new_col.str.cat(df[col_name].astype(str), sep=step.separator)
    return df.assign(**{step.new_column_name: new_col})



from typing import List, Literal, Union

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import TemplatedVariable


class PivotStep(BaseStep):
    name: Literal['pivot'] = 'pivot'
    index: List[str]
    column_to_pivot: str
    value_column: str
    agg_function: Literal['sum', 'avg', 'count', 'min', 'max']


class PivotStepWithVariable(PivotStep, StepWithVariablesMixin):
    index: Union[TemplatedVariable, List[TemplatedVariable]]



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import RankStep


def execute_rank(
    step: RankStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column_name = step.new_column_name or f'{step.value_col}_RANK'
    rank_method = 'min' if step.method == 'standard' else step.method
    ascending = step.order == 'asc'
    if step.groupby:
        serie = df.groupby(step.groupby, dropna=False)[step.value_col]
    else:
        serie = df[step.value_col]
    rank_serie = serie.rank(method=rank_method, ascending=ascending)
    return df.assign(**{new_column_name: rank_serie}).sort_values(new_column_name)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import PercentageStep


def translate_percentage(step: PercentageStep) -> List[MongoStep]:
    return [
        {
            '$group': {
                '_id': {g: f'${g}' for g in step.group} if step.group else None,
                '_vqbAppArray': {'$push': '$$ROOT'},
                '_vqbTotalDenum': {'$sum': f'${step.column }'},
            },
        },
        {'$unwind': '$_vqbAppArray'},
        {
            '$project': {
                step.new_column_name
                or f'{step.column}_PCT': {
                    '$cond': [
                        {'$eq': ['$_vqbTotalDenum', 0]},
                        None,
                        {'$divide': [f'$_vqbAppArray.{step.column}', '$_vqbTotalDenum']},
                    ],
                },
                '_vqbAppArray': 1,  # we need to keep track of this key for the next operation
            },
        },
        # Line below: Keep all columns that were not used in computation, 'stored' in _vqbAppArray
        {'$replaceRoot': {'newRoot': {'$mergeObjects': ['$_vqbAppArray', '$$ROOT']}}},
        {'$project': {'_vqbAppArray': 0}},  # We do not want to keep that column at the end
    ]



import datetime
from re import sub
from typing import Any, Dict, List, Union

from weaverbird.backends.mongo_translator.steps.filter import (
    translate_relative_date,
    truncate_to_day,
)
from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.conditions import (
    ConditionComboAnd,
    ConditionComboOr,
    MatchCondition,
    SimpleCondition,
)
from weaverbird.pipeline.dates import RelativeDate


def column_to_user_variable(col_name: str) -> str:
    """User variable names can contain the ascii characters [_a-zA-Z0-9] and any non-ascii character."""
    col_name_without_invalid_chars = sub(r'/[^_a-zA-Z0-9]/g', '_', col_name)
    return f'vqb_{col_name_without_invalid_chars}'


class UnsupportedOperatorError(Exception):
    """Raised when condition uses an unsupported operator"""


def build_cond_expression(
    cond: Union[SimpleCondition, ConditionComboOr, ConditionComboAnd]
) -> MongoStep:
    operator_mapping = {
        'eq': '$eq',
        'ne': '$ne',
        'lt': '$lt',
        'le': '$lte',
        'gt': '$gt',
        'ge': '$gte',
        'in': '$in',
        'nin': '$in',
        'isnull': '$eq',
        'notnull': '$ne',
        'matches': '$regexMatch',
        'notmatches': '$regexMatch',
        'from': '$gte',
        'until': '$lte',
    }
    unsupported_operators: List = []

    if isinstance(cond, ConditionComboAnd):
        return build_and_expression(cond)
    if isinstance(cond, ConditionComboOr):
        return {'$and': [build_cond_expression(elem) for elem in cond.or_]}
    if cond.operator in unsupported_operators:
        raise UnsupportedOperatorError(f'Unsupported operator ${cond.operator} in conditions')

    if cond.operator == 'matches' or cond.operator == 'notmatches':
        cond_expression = build_matches_expression(cond)
        return cond_expression

    else:
        if cond.operator == 'notnull' or cond.operator == 'isnull':
            return {operator_mapping[cond.operator]: [f'${cond.column}', None]}

        else:
            cond_expression = {operator_mapping[cond.operator]: [f'${cond.column}', cond.value]}
            if cond.operator == 'nin':
                cond_expression = {'$not': cond_expression}

            cond_expression = build_dates_expressions(cond, cond_expression, operator_mapping)
            return cond_expression


def build_dates_expressions(
    cond: SimpleCondition, cond_expression: Dict[str, Any], operator_mapping: Dict[str, str]
):
    if cond.operator == 'until':
        if isinstance(cond.value, datetime.datetime):
            cond_expression[operator_mapping[cond.operator]][1] = [
                datetime.datetime(
                    day=cond.value.day, month=cond.value.month, year=cond.value.month
                ).replace(hour=23, minute=59, second=59, microsecond=999999)
            ]
    if cond.operator == 'from' or cond.operator == 'until':
        cond_expression = {
            operator_mapping[cond.operator]: [
                truncate_to_day(f'${cond.column}'),
                truncate_to_day(
                    translate_relative_date(cond.value)
                    if isinstance(cond.value, RelativeDate)
                    else cond.value
                ),
            ]
        }
    return cond_expression


def build_matches_expression(cond: MatchCondition):
    cond_expression: MongoStep = {'$regexMatch': {'input': f'${cond.column}', 'regex': cond.value}}
    if cond.operator == 'notmatches':
        cond_expression = {'$not': cond_expression}
    return cond_expression


def build_and_expression(cond: ConditionComboAnd):
    if len(cond.and_) == 1:
        return build_cond_expression(cond.and_[0])
    else:
        return {'$and': [build_cond_expression(elem) for elem in cond.and_]}



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class CompareTextStep(BaseStep):
    name: Literal['comparetext'] = 'comparetext'
    new_column_name: ColumnName = Field(alias='newColumnName')
    str_col_1: ColumnName = Field(alias='strCol1')
    str_col_2: ColumnName = Field(alias='strCol2')


class CompareTextStepWithVariables(CompareTextStep, StepWithVariablesMixin):
    str_col_1: TemplatedVariable = Field(alias='strCol1')
    str_col_2: TemplatedVariable = Field(alias='strCol2')



from enum import Enum


class SQLDialect(str, Enum):
    ATHENA = "athena"
    GOOGLEBIGQUERY = "googlebigquery"
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    REDSHIFT = "redshift"
    SNOWFLAKE = "snowflake"



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import DurationStep

_SECOND = 1
_MINUTE = _SECOND * 60
_HOUR = _MINUTE * 60
_DAY = _HOUR * 24

DURATIONS_IN_SECOND = {'seconds': _SECOND, 'minutes': _MINUTE, 'hours': _HOUR, 'days': _DAY}


def execute_duration(
    step: DurationStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    duration_serie_in_seconds = (df[step.end_date_column] - df[step.start_date_column]).astype(
        'timedelta64[s]'
    )
    duration_serie_in_given_unit = duration_serie_in_seconds / DURATIONS_IN_SECOND[step.duration_in]
    return df.assign(**{step.new_column_name: duration_serie_in_given_unit})



from itertools import chain as ichain
from typing import Any, Literal

from pydantic import validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName

from .aggregate import Aggregation


class DissolveStep(BaseStep):
    name: Literal['dissolve'] = 'dissolve'
    groups: list[ColumnName]
    aggregations: list[Aggregation] = []
    include_nulls: bool = False

    @staticmethod
    def _ensure_unique_and_non_empty(values: list[Any], col_name: str) -> None:
        assert all(len(v) for v in values), f"all values in '{col_name}' must be non-empty"
        assert len(values) == len(set(values)), f"all values in '{col_name}' must be unique"

    @validator('groups')
    def _len_validator(cls, values: list) -> list:
        assert len(values) > 0, "list must contain at least one element"
        return values

    @validator('aggregations')
    def _validate_aggregations(cls, values: list[Aggregation]) -> list[Aggregation]:
        if len(values) < 1:
            return values

        for agg in values:
            assert len(agg.columns) == 1, "aggregations can only contain a single column"
            assert len(agg.new_columns) == 1, "aggregations can only contain a single new column"

        cls._ensure_unique_and_non_empty(
            list(ichain.from_iterable(agg.columns for agg in values)), 'columns'
        )
        cls._ensure_unique_and_non_empty(
            list(ichain.from_iterable(agg.new_columns for agg in values)), 'new_columns'
        )
        return values



from typing import Dict, List, Union

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ConcatenateStep


def _convert_to_type(input: Union[str, dict], type_: str) -> Dict[str, Union[str, dict]]:
    return {'$convert': {'input': input, 'to': type_}}


def translate_concatenate(step: ConcatenateStep) -> List[MongoStep]:
    # NOTE: Translated from "../src/lib/translators/mongo.ts"'s
    # concatenate() step. The difference is that we're using mongo's
    # $convert operator here, which requires mongo>=4.0
    concat_block: List[Union[dict, str]] = [_convert_to_type(f'${step.columns[0]}', 'string')]
    for col in step.columns[1:]:
        concat_block += [step.separator, _convert_to_type(f'${col}', 'string')]

    return [{'$addFields': {step.new_column_name: {'$concat': concat_block}}}]



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import PivotStep


def execute_pivot(
    step: PivotStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    pivoted_df = df.pivot_table(
        values=step.value_column,
        index=step.index,
        columns=step.column_to_pivot,
        aggfunc='mean' if step.agg_function == 'avg' else step.agg_function,
    ).reset_index()
    pivoted_df.columns.name = None
    return pivoted_df



from typing import Literal, Optional

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class SubstringStep(BaseStep):
    name: Literal['substring'] = 'substring'
    column: ColumnName
    new_column_name: Optional[ColumnName] = Field(alias='newColumnName')
    start_index: int
    end_index: int



from pandas import DataFrame

from weaverbird.backends.pandas_executor.geo_utils import df_to_geodf
from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import SimplifyStep


def execute_simplify(
    step: SimplifyStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    geo_df = df_to_geodf(df)
    # WARNING: potentially mutating
    # Simplify returns a GeoSeries, so we need to reassign here.
    #
    # Also, we don't want to provide preserve_topoly to the user, as results
    # are too imprecise (some result in empty polygons, see:
    # https://geopandas.org/en/stable/docs/reference/api/geopandas.GeoSeries.simplify.html)
    geo_df.geometry = geo_df.simplify(tolerance=step.tolerance, preserve_topology=True)
    return DataFrame(geo_df)



from typing import List, Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class TrimStep(BaseStep):
    name: Literal['trim'] = 'trim'
    columns: List[ColumnName]



from typing import List, Union

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DateExtractStep


def _truncate_date_to_day(expr: Union[dict, str]) -> MongoStep:
    return {
        '$dateTrunc': {
            'unit': 'day',
            'date': expr,
        },
    }


def _extract_quarter(step: DateExtractStep) -> MongoStep:
    return {
        '$switch': {
            'branches': [
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 1]}, 'then': 1},
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 2]}, 'then': 2},
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 3]}, 'then': 3},
            ],
            'default': 4,
        },
    }


def _extract_first_day_of_year(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {'year': {'$year': f'${step.column}'}, 'month': 1, 'day': 1},
    }


def _extract_first_day_of_month(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {
            'year': {'$year': f'${step.column}'},
            'month': {'$month': f'${step.column}'},
            'day': 1,
        },
    }


def _extract_first_day_of_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (dayOfWeek - 1)
            '$subtract': [
                f'${step.column}',
                {
                    '$multiply': [
                        {'$subtract': [{'$dayOfWeek': f'${step.column}'}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_first_day_of_quarter(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {
            'year': {'$year': f'${step.column}'},
            'month': {
                '$switch': {
                    'branches': [
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 1]},
                            'then': 1,
                        },
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 2]},
                            'then': 4,
                        },
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 3]},
                            'then': 7,
                        },
                    ],
                    'default': 10,
                },
            },
            'day': 1,
        },
    }


def _extract_first_day_of_iso_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (isoDayOfWeek - 1)
            '$subtract': [
                f'${step.column}',
                {
                    '$multiply': [
                        {'$subtract': [{'$isoDayOfWeek': f'${step.column}'}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_previous_day(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date 1 day in milliseconds
            '$subtract': [f'${step.column}', 24 * 60 * 60 * 1000],
        }
    )


def _extract_first_day_of_previous_year(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {
            'year': {'$subtract': [{'$year': f'${step.column}'}, 1]},
            'month': 1,
            'day': 1,
        },
    }


def _extract_first_day_of_previous_month(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {
            'year': {
                '$cond': [
                    {'$eq': [{'$month': f'${step.column}'}, 1]},
                    {'$subtract': [{'$year': f'${step.column}'}, 1]},
                    {'$year': f'${step.column}'},
                ],
            },
            'month': {
                '$cond': [
                    {'$eq': [{'$month': f'${step.column}'}, 1]},
                    12,
                    {'$subtract': [{'$month': f'${step.column}'}, 1]},
                ],
            },
            'day': 1,
        },
    }


def _extract_first_day_of_previous_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (dayOfWeek - 1)
            '$subtract': [
                {'$subtract': [f'${step.column}', 7 * 24 * 60 * 60 * 1000]},
                {
                    '$multiply': [
                        {'$subtract': [{'$dayOfWeek': f'${step.column}'}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_first_day_of_previous_quarter(step: DateExtractStep) -> MongoStep:
    return {
        '$dateFromParts': {
            'year': {
                '$cond': [
                    {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 1]},
                    {'$subtract': [{'$year': f'${step.column}'}, 1]},
                    {'$year': f'${step.column}'},
                ],
            },
            'month': {
                '$switch': {
                    'branches': [
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 1]},
                            'then': 10,
                        },
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 2]},
                            'then': 1,
                        },
                        {
                            'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 3]},
                            'then': 4,
                        },
                    ],
                    'default': 7,
                },
            },
            'day': 1,
        },
    }


def _extract_first_day_of_previous_iso_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (isoDayOfWeek - 1)
            '$subtract': [
                {'$subtract': [f'${step.column}', 7 * 24 * 60 * 60 * 1000]},
                {
                    '$multiply': [
                        {'$subtract': [{'$isoDayOfWeek': f'${step.column}'}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_previous_year(step: DateExtractStep) -> MongoStep:
    return {
        '$subtract': [{'$year': f'${step.column}'}, 1],
    }


def _extract_previous_month(step: DateExtractStep) -> MongoStep:
    return {
        '$cond': [
            {'$eq': [{'$month': f'${step.column}'}, 1]},
            12,
            {'$subtract': [{'$month': f'${step.column}'}, 1]},
        ],
    }


def _extract_previous_week(step: DateExtractStep) -> MongoStep:
    return {
        # We subtract to the target date 7 days in milliseconds
        '$week': {'$subtract': [f'${step.column}', 7 * 24 * 60 * 60 * 1000]},
    }


def _extract_previous_quarter(step: DateExtractStep) -> MongoStep:
    return {
        '$switch': {
            'branches': [
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 1]}, 'then': 4},
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 2]}, 'then': 1},
                {'case': {'$lte': [{'$divide': [{'$month': f'${step.column}'}, 3]}, 3]}, 'then': 2},
            ],
            'default': 3,
        },
    }


def _extract_previous_iso_week(step: DateExtractStep) -> MongoStep:
    return {
        # We subtract to the target date 7 days in milliseconds
        '$isoWeek': {'$subtract': [f'${step.column}', 7 * 24 * 60 * 60 * 1000]},
    }


_ADVANCED_DATE_EXTRACT_MAP = {
    'quarter': _extract_quarter,
    'firstDayOfYear': _extract_first_day_of_year,
    'firstDayOfMonth': _extract_first_day_of_month,
    'firstDayOfWeek': _extract_first_day_of_week,
    'firstDayOfQuarter': _extract_first_day_of_quarter,
    'firstDayOfIsoWeek': _extract_first_day_of_iso_week,
    'previousDay': _extract_previous_day,
    'firstDayOfPreviousYear': _extract_first_day_of_previous_year,
    'firstDayOfPreviousMonth': _extract_first_day_of_previous_month,
    'firstDayOfPreviousWeek': _extract_first_day_of_previous_week,
    'firstDayOfPreviousQuarter': _extract_first_day_of_previous_quarter,
    'firstDayOfPreviousIsoWeek': _extract_first_day_of_previous_iso_week,
    'previousYear': _extract_previous_year,
    'previousMonth': _extract_previous_month,
    'previousWeek': _extract_previous_week,
    'previousQuarter': _extract_previous_quarter,
    'previousIsoWeek': _extract_previous_iso_week,
}

_DATE_EXTRACT_MAP = {
    'year': '$year',
    'month': '$month',
    'day': '$dayOfMonth',
    'week': '$week',
    'dayOfYear': '$dayOfYear',
    'dayOfWeek': '$dayOfWeek',
    'isoYear': '$isoWeekYear',
    'isoWeek': '$isoWeek',
    'isoDayOfWeek': '$isoDayOfWeek',
    'hour': '$hour',
    'minutes': '$minute',
    'seconds': '$second',
    'milliseconds': '$millisecond',
}


def translate_date_extract(step: DateExtractStep) -> List[MongoStep]:
    new_columns = []
    add_fields = {}

    date_info: list
    # For retrocompatibility
    if step.operation:
        date_info = [step.operation] if step.operation else step.date_info
        new_columns = [
            step.new_column_name if step.new_column_name else f"{step.column}_{step.operation}"
        ]
    else:
        date_info = step.date_info.copy()
        new_columns = step.new_columns.copy()

    for i, d in enumerate(date_info):
        if d in _ADVANCED_DATE_EXTRACT_MAP:
            add_fields[new_columns[i]] = _ADVANCED_DATE_EXTRACT_MAP[d](step)
        else:
            add_fields[new_columns[i]] = {
                _DATE_EXTRACT_MAP[d]: f"${step.column}",
            }

    return [{'$addFields': add_fields}]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['weaverbird',
 'weaverbird.backends',
 'weaverbird.backends.mongo_translator',
 'weaverbird.backends.mongo_translator.steps',
 'weaverbird.backends.pandas_executor',
 'weaverbird.backends.pandas_executor.steps',
 'weaverbird.backends.pandas_executor.steps.utils',
 'weaverbird.backends.pypika_translator',
 'weaverbird.backends.pypika_translator.translators',
 'weaverbird.backends.sql_translator',
 'weaverbird.backends.sql_translator.steps',
 'weaverbird.backends.sql_translator.steps.utils',
 'weaverbird.pipeline',
 'weaverbird.pipeline.steps',
 'weaverbird.pipeline.steps.utils',
 'weaverbird.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyPika>=0.48.9,<0.49.0',
 'geopandas>=0.10.2,<0.11.0',
 'numexpr>=2.8.1,<3.0.0',
 'pandas>=1.2.5,<2.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'redshift-connector>=2.0.907,<3.0.0',
 'sqlalchemy-redshift>=0.8.9,<0.9.0',
 'typing_extensions>=4.2,<5.0']

extras_require = \
{'playground': ['quart>=0.17,<0.18',
                'hypercorn>=0.13,<0.14',
                'pymongo[srv,tls]>=3.12.0',
                'snowflake-connector-python[pandas]>=2.7.8,<3.0.0']}

setup_kwargs = {
    'name': 'weaverbird',
    'version': '0.15.0',
    'description': 'A visual data pipeline builder with various backends',
    'long_description': "# weaverbird python package\n\nSee [docs about purpose and usage](../docs/_docs/tech/python-package.md).\n\n## Development\n\nWe use [poetry](https://python-poetry.org/) for managing dependencies.\n\nMain commands are available through `make`:\n\n    make install # Install dependecies\n\n    make format # Fix formatting issues using black and isort\n    make lint # Execute various checks\n\n    make build # Build the project prior to publication\n    make upload # Publish on pypi\n\n    make test # Execute the test suite and produce reports\n    /!\\ To run Snowflake's e2e tests, the password needs to be exported to env variables\n    as such: export SNOWFLAKE_PASSWORD='XXXXXXXXXXX'. This password is available in lastpass (user: toucan_test)\n\n### Playground server\n\nSee `playground.py`. It provides a very simple server to test the module.\n",
    'author': 'Toucan Toco',
    'author_email': 'dev@toucantoco.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4',
}


setup(**setup_kwargs)



from typing import Mapping, Sequence

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.translators import ALL_TRANSLATORS
from weaverbird.pipeline import PipelineWithVariables


def translate_pipeline(
    *,
    sql_dialect: SQLDialect,
    pipeline: PipelineWithVariables,
    tables_columns: Mapping[str, Sequence[str]],
    db_schema: str | None = None,
) -> str:
    translator_cls = ALL_TRANSLATORS[sql_dialect]
    translator = translator_cls(
        tables_columns=tables_columns,
        db_schema=db_schema,
    )
    return translator.get_query_str(steps=pipeline.steps)



from pypika.dialects import SnowflakeQuery

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import DataTypeMapping, SQLTranslator


class SnowflakeTranslator(SQLTranslator):
    DIALECT = SQLDialect.SNOWFLAKE
    QUERY_CLS = SnowflakeQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.TO_DATE


SQLTranslator.register(SnowflakeTranslator)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import TrimStep


def translate_trim(step: TrimStep) -> List[MongoStep]:
    add_fields = {x: {'$trim': {'input': f'${x}'}} for x in step.columns}

    return [{'$addFields': add_fields}]



from numpy.ma import logical_and, logical_or
from pandas import DataFrame, Series
from pandas.tseries.offsets import DateOffset

from weaverbird.backends.pandas_executor.steps.utils.dates import evaluate_relative_date
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    Condition,
    ConditionComboAnd,
    ConditionComboOr,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.dates import RelativeDate


def apply_condition(condition: Condition, df: DataFrame) -> Series:
    """
    Returns a boolean Series, that will be used to filter a DataFrame.

    Example:
        filter = apply_condition(condition, df)
        df[filter]  # The filtered DataFrame
    """

    if isinstance(condition, ComparisonCondition):
        return getattr(df[condition.column], condition.operator)(condition.value)
    elif isinstance(condition, InclusionCondition):
        f = df[condition.column].isin(condition.value)
        if condition.operator.startswith('n'):
            return ~f
        else:
            return f
    elif isinstance(condition, NullCondition):
        f = df[condition.column].isnull()
        if condition.operator.startswith('not'):
            return ~f
        else:
            return f
    elif isinstance(condition, MatchCondition):
        f = df[condition.column].str.contains(condition.value)
        if condition.operator.startswith('not'):
            return ~f
        else:
            return f

    elif isinstance(condition, DateBoundCondition):
        if condition.operator == 'until':
            comparison_method = 'le'
        elif condition.operator == 'from':
            comparison_method = 'ge'
        else:
            raise NotImplementedError

        value = condition.value
        if isinstance(value, RelativeDate):
            value = evaluate_relative_date(value)

        # Remove time info from the column to filter on
        column_without_time = df[condition.column] - DateOffset(
            hour=0, minute=0, second=0, microsecond=0, nanosecond=0
        )
        # Do the same with the value to compare it to
        value_without_time = value - DateOffset(
            hour=0, minute=0, second=0, microsecond=0, nanosecond=0
        )

        return getattr(column_without_time, comparison_method)(value_without_time)

    elif isinstance(condition, ConditionComboAnd):
        return logical_and.reduce([apply_condition(c, df) for c in condition.and_])
    elif isinstance(condition, ConditionComboOr):
        return logical_or.reduce([apply_condition(c, df) for c in condition.or_])
    else:
        raise NotImplementedError



from typing import List, Literal, Optional, Tuple, Union

from pydantic import Field, root_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class CumSumStep(BaseStep):
    name: Literal['cumsum'] = 'cumsum'
    to_cumsum: List[Tuple[str, Optional[str]]] = Field(..., alias='toCumSum')
    reference_column: ColumnName = Field(..., alias='referenceColumn')
    groupby: Optional[List[ColumnName]]

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'valueColumn' in values:
            values['value_column'] = values.pop('valueColumn')
        if 'newColumn' in values:
            values['new_column'] = values.pop('newColumn')

        if 'value_column' in values:
            values['to_cumsum'] = [(values.pop('value_column'), values.pop('new_column', None))]
        return values


class CumSumStepWithVariable(CumSumStep, StepWithVariablesMixin):
    to_cumsum: Union[TemplatedVariable, List[Tuple[TemplatedVariable, TemplatedVariable]]] = Field(
        ..., alias='toCumSum'
    )



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import UppercaseStep


def translate_uppercase(step: UppercaseStep) -> List[MongoStep]:
    return [{'$addFields': {step.column: {'$toUpper': f'${step.column}'}}}]
